/*
 * kernel.h
 *
 * Central routines to drive the digital logic simulator.
 * -eichin 19880306.0104EST
 */

/*
 * freecell
 * adds the args to the freelist.
 */
freecell(cell)
     list cell;
{
}

/*
 * wire_new
 * creates a new wire, ready to add stuff to.
 */

Wire wire_new()
{
}

wire_add_depend(wire, dep)
     Wire wire;
     Depend dep;
{
}
  
wire_add_event(wire, event)
     Wire wire;
     Event event;
{
}


/*
 * depend_new
 * creates a new dependency
 */

Depend depend_new(func, arglist)
{
}

/*
 * event_new
 * creates a new event, at a given time.
 */

typedef enum _Event_val {
  LOGIC_HI, LOGIC_LOW, 
  EARLY_HI, EARLY_LOW,
  TRISTATE, CONTENTION,
  UNDEFINED, FILE_FETCH,
} Event_val;

typedef long Time;

typedef Time Event_time;

typedef struct _Event {
  Event_val	val;
  Event_time	time;
} Event;

typedef struct _Timeline {
  struct _Timeline	*next_line;
  Event			*this_event;
} Timeline;

typedef struct _Args {
  struct _Args		*next_arg;
  struct _Wire		*this_wire;
} Args;

typedef struct _Function {
  struct _Args		*next_args;
  struct _Function	*this_func;
} Function;


typedef struct _Depend {
  struct _Depend	*next_depend;
  Function		*this_func;
}

typedef struct _Wire {
  Depend	*depend;
  Timeline	*timeline;
} Wire;

typedef struct _Wire_loop_el {
  struct _Wire_loop_el	*next_loop;
  Wire			*this_wire; /* if null, check file input. */
} Wire_loop_el;




Event event_new(val, time)
     Event_val val;
     Event_time time;
{
  /*
   * events should be allocated out of their own pool, though they
   * may be the same size as the other allocated objects, since
   * 	1. the struct size may be architecture specific
   * 	2. they get disposed in their own special way
   */
  
}

Time event_get_time(event)
     Event event;
{
  if(event.val != FILE_FETCH)
    {
      return(event.time);
    }
  else
    {
      return(NOT_A_TIME);
    }
}

Event_val event_get_val(event)
     Event event;
{
  return(event.val);
}

#define local

local Event event_set_val(event, val, time)
     Event event;
     Event_val val;
     Event_time time;
{
  event.val = val;
  event.time = time;
}

wire_append_event(wire, val, time)
     Wire wire;
     Event_val val;
     Event_time time;
{
  event_new(val, time);

}

/*
 * The timeline of the wire should have the following parts:
 * 	a. current state
 * 	b. past state (optional)
 * 	c. future states
 * 		i. pointer to next possible time
 * 		ii. pointer to last known time
 * 	d. contention records
 */

Time timeline_next_change(timeline)
     Timeline timeline;
{
  Time t;
  if ( FETCH_FILE == event_get_val(timeline_get_event(timeline)) )
    {
      timeline_fetch_file(timeline);
    }
  return(event_get_time(timeline_get_event(timeline)));
}

#define error_line(mod,arg) do {					\
  fprintf(stderr, "Error(%s):%s, line %d\n", mod, __FILE__, __LINE__);	\
  fprintf arg;								\
  abort();								\
} while(0)

timeline_fetch_file()
{
  /*
   * here we update the timeline from a file. Specifications are in
   * the Tools-11 document. [eichin:19880307.0335EST]
   */

  /*
   * The data are in the form of assignments and waits. This implies a
   * reader loop with a current timestamp which reads and sets until
   * it reaches a ``wait'' statement, and then increments the timestamp.
   * Line names are looked up in a name table, the values are added
   * directly to the wires. The values are mapped as follows:
   * 	1	LOGIC_HI
   * 	0	LOGIC_LOW
   * 	U	UNDEFINED
   * 	C	CONTENTION
   * 	X	TRISTATE
   * ``wait'' may not have an argument, in which case logic values are
   * intended to settle; this should probably be scheduled as a wait
   * of a single tick. [eichin:19880307.0349EST]
   */

  /*
   * In order to make this loop read the input stream effectively,
   * without `flooding', we use non-blocking reads to get as much as
   * we can without waiting. This will be isolated so that for systems
   * that this doesn't port well to, we can simply read the next five
   * waits, or something like that. [eichin:19880307.0400EST]
   */

  /*
   * This should get called from the main scheduling loop. Since the
   * file specifies by name which wires are changing, it can be
   * checked independently of the individual wires. [eichin:19880309.1109EST]
   */
  Time now = current_time();
  
  
  while(input_data_avail(fid))
    {
      char *cmd_s;
      cmd_s = input_get_cmd(fid);
      if(is_wait(cmd_s))
	{
	  int arg;
	  char *arg_s = input_get_arg(fid);
	  if(!arg_s) {
	    arg = 1 TICK;
	  } else {
	    arg = atoi(arg_s) UNITS;
	  }
	  now += arg;
	}
      else
	{
	  Wire wire;
	  if(!(wire = lookup_wire_name(cmd_s)))
	    {
	      error_line(ERR_TIMEVAL, (stderr, "No such wire <%s>\n", cmd_s));
	    }
	  wire_append_event(wire, arg, now);
	}
    }
	    
	      
  
}

set_noblock(fd)
     int fd;
{
  /*
   * set it non-blocking.  This is so we can read data when blocked
   * for writing in the library.
   */
#ifdef FIOSNBIO
  {
    int arg = 1;
    ioctl(fd, FIOSNBIO, &arg);
  }
#else
  (void) fcntl(fd, F_SETFL, FNDELAY);
#endif /* FIOSNBIO */
}



void changed_connection(list cnn); /* iterate over conn's procedures */

int search_all(list cnnlist);	/* scan list of all connections for
				 * earliest event
				 */

void scan_all(list cnnlist; int tmin); /* do the next time tick */

void run(list connlist);	/* complete simulation */

list newconn(list init_value);	/* create a connection */
