/*
 * ropes.c
 * A rope is a thick string.
 */

#include "decls.h"

void rop_debug();


rope *ropcat(r1, r2)
     rope *r1, *r2;
{
  register rope *rs = r1;

  while( rs->typ != END )	/* scan for the end of rope... */
    {
      if ( rs->typ == NEXT )	/* don't stumble on NEXTs... */
	{
	  rs = rs->ind.next;
	}
      else
	{
	  rs++;
	}
    }
  rs->typ = NEXT;		/* set link to r2 */
  rs->ind.next = r2;

  return(r1);
}

void ropappend(r, re)
     rope *r;
     rope re;
{
  register rope *rs = r;

  while( rs->typ != END )	/* scan for end of rope... */
    {
      if ( rs->typ == NEXT )	/* don't stumble on NEXTs... */
	{
	  rs = rs->ind.next;
	  r = rs;		/* we only need the size on the last one */
	}
      else
	{
	  rs++;
	}
    }

  if ( (rs - r + 1) < ROPE_SIZE ) /* we have at least one left, bump the end */
    {
      *rs = re;
      rs++;
      rs->typ = END;
    }
  else				/* END in final slot, make a NEXT and jump */
    {
      rs->typ = NEXT;
      rs->ind.next = ropinit();
      rs = rs->ind.next;
      *rs = re;
      rs++;
      rs->typ = END;
    }
}

rope *rop_cdr(r)
     rope *r;
{
  register rope *rs = r;
  int i = 0;

  /*
   * Scan for the first real thing to delete it.
   */
  while((rs->tag == NULL) && (i++ < ROPE_SIZE))
    {
      rs++;
    }
  
  
  if(rs->typ == NEXT)
    {
      rs = rs->val;
    }
  ropfree(r);

  return(rs);
}

rope *ropinit()
{
  rope *rs;

  /* figure out whether to malloc or not...or use a freelist? :-) */
  rs = (rope *)malloc(ROPE_SIZE * sizeof(rope));
  rs->typ = END;
  return(rs);
}

void ropfree(r)
     rope *r;
{
  rope *rs = r;

  while(r->typ != END)
    {
      if (r->typ == NEXT)
	{
	  r = r->ind.next;
	  free(rs);
	  rs = r;
	}
      else if (r->typ == ROPE)
	{
	  ropfree(r->ind.rope);
	  r++;
	}
      else
	{
	  relfree(*r);
	  r++;
	}
    }
  free(rs);
}

