/*
 * dff.c
 * a few pieces for a running simulator.
 * [eichin:19880406.1107EST]
 */

#include <msimdecl.h>

#define dff_tsetup 5
#define dff_tprop0 11
#define dff_tprop1 13

/*
 * z, z' = dff(clk, in, preset, clear)
 */

/*
 * Table from page 5-22, TI databook, for 7474:
 * ========================================
 * inputs			outputs
 * ----------------------------------------
 * preset clear	clock	D	Q	Q'
 * ========================================
 * L	H	X	X	H	L
 * H	L	X	X	L	H
 * L	L	X	X	H*	H*
 * H	H	^	H	H	L
 * H	H	^	L	L	H
 * H	H	L	X	Q0	Q'0
 * ========================================
 * *unstable.
 * [eichin:19880406.1148EST]
 */
dff_func(dep)
     Generics dep;
{
  Wire outz, outznot;
  Timeline outz_time, outznot_time;
  Wire inclk, inin, inpreset, inclear;
  Timeline inclk_time, inin_time, inpreset_time, inclear_time;
  
  nextwire(dep, outz, outz_time);
  nextwire(dep, outznot, outznot_time);

  readnexttime(dep, inclk, inclk_time);
  readnexttime(dep, inin, inin_time);
  readnexttime(dep, inpreset, inpreset_time);
  readnexttime(dep, inclear, inclear_time);

  /* was it preset? */
  if(inpreset_time[0].tag.EVENT == LOGIC_LOW)
    {
      add_event(outz_time, LOGIC_HI, globaltime + dff_tprop0);
      add_event(outznot_time, LOGIC_LOW, globaltime + dff_tprop0);
    }
  /* was it clear? */
  else if(inclear_time[0].tag.EVENT == LOGIC_LOW)
    {
      add_event(outz_time, LOGIC_LOW, globaltime + dff_tprop0);
      add_event(outznot_time, LOGIC_HI, globaltime + dff_tprop0);
    }
  /* was it a clock tick? */
  else if((inclk_time[0].data.EVENT == globaltime)
	  && (inclk_time[0].tag.EVENT == LOGIC_HI))
    {
      /* ok, was the input stable? */
      if(inin_time[0].data.EVENT <= globaltime - dff_tsetup)
	{
	  /* Ok, changing now... */
	  add_event(outz_time,
		    inin_time[0].tag.EVENT,
		    inclk_time[0].data.EVENT + dff_tprop1);
	}
    }
}
